/*-
 * Copyright (c) 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* for strdup() */
#define _POSIX_C_SOURCE 200809L

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cache/cache.h"
#include "vcl.h"
#include "vre.h"
#include "vbm.h"
#include "cache/cache_director.h"

#include "vcc_if.h"
#include "patricia.h"

#define VFAIL(ctx, fmt, ...) \
	VRT_fail((ctx), "vmod selector failure: " fmt, __VA_ARGS__)

#define VERR(ctx, fmt, ...) \
	VSLb((ctx)->vsl, SLT_VCL_Error, "vmod selector error: " fmt,	\
	     __VA_ARGS__)

#define VERRNOMEM(ctx, fmt, ...) \
	VERR((ctx), "out of space: " fmt,  __VA_ARGS__)

struct entry {
	unsigned	magic;
#define VMOD_SELECTOR_ENTRY_MAGIC 0x733dbe63
	char		*string;
	VCL_BACKEND	backend;
	vre_t		*re;
};

enum bitmap_e {
	STRING = 0,
	BACKEND,
	REGEX,
	__MAX_BITMAP,
};

struct bitmaps {
	unsigned	magic;
#define VMOD_SELECTOR_BITMAPS_MAGIC 0x5b17093f
	struct vbitmap	*bitmaps[__MAX_BITMAP];
};

struct vmod_selector_set {
	unsigned	magic;
#define VMOD_SELECTOR_SET_MAGIC 0x838979ef
	struct entry	**table;
	char		**members;
	char		**lomembers;
	struct pt_y	*origo;
	char		*vcl_name;
	struct bitmaps	*bitmaps;
	unsigned	nmembers;
	unsigned	nentries;
	VCL_BOOL	case_sensitive;
};

/* Event function */

int
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
        (void) ctx;
        (void) priv;

        if (e == VCL_EVENT_LOAD && !PT_Inited())
		PT_Init();
        return 0;
}

/* Object regex */

VCL_VOID
vmod_set__init(VRT_CTX, struct vmod_selector_set **setp, const char *vcl_name,
	       VCL_BOOL case_sensitive)
{
	struct vmod_selector_set *set;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(setp);
	AZ(*setp);
	AN(vcl_name);
	
	ALLOC_OBJ(set, VMOD_SELECTOR_SET_MAGIC);
	AN(set);
	*setp = set;
	set->vcl_name = strdup(vcl_name);
	AN(set->vcl_name);
	set->case_sensitive = case_sensitive;

	ALLOC_OBJ(set->bitmaps, VMOD_SELECTOR_BITMAPS_MAGIC);
	AN(set->bitmaps);
	for (int i = 0; i < __MAX_BITMAP; i++) {
		set->bitmaps->bitmaps[i] = vbit_new(0);
		AN(set->bitmaps->bitmaps[i]);
	}

	AZ(set->table);
	AZ(set->members);
	AZ(set->origo);
	AZ(set->nentries);
}

VCL_VOID
vmod_set__fini(struct vmod_selector_set **setp)
{
	struct vmod_selector_set *set;

	if (setp == NULL || *setp == NULL)
		return;
	CHECK_OBJ(*setp, VMOD_SELECTOR_SET_MAGIC);
	set = *setp;
	*setp = NULL;
	PT_Free(set->origo);
	for (unsigned i = 0; i < set->nmembers; i++)
		free(set->members[i]);
	for (unsigned i = 0; i < set->nentries; i++) {
		struct entry *entry = set->table[i];
		CHECK_OBJ_NOTNULL(entry, VMOD_SELECTOR_ENTRY_MAGIC);
		free(entry->string);
		if (entry->re != NULL)
			VRE_free(&entry->re);
		FREE_OBJ(entry);
	}
	for (int i = 0; i < __MAX_BITMAP; i++)
		vbit_destroy(set->bitmaps->bitmaps[i]);
	FREE_OBJ(set->bitmaps);

	free(set->members);
	free(set->table);
	free(set->vcl_name);
	FREE_OBJ(set);
}

static inline void
set_added(struct vmod_selector_set *set, unsigned idx, enum bitmap_e bitmap)
{
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);
	CHECK_OBJ_NOTNULL(set->bitmaps, VMOD_SELECTOR_BITMAPS_MAGIC);
	AN(set->bitmaps->bitmaps[bitmap]);

	vbit_set(set->bitmaps->bitmaps[bitmap], idx);
}

static inline int
is_added(const struct vmod_selector_set *set, unsigned idx,
	 enum bitmap_e bitmap)
{
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);
	CHECK_OBJ_NOTNULL(set->bitmaps, VMOD_SELECTOR_BITMAPS_MAGIC);
	AN(set->bitmaps->bitmaps[bitmap]);

	return (vbit_test(set->bitmaps->bitmaps[bitmap], idx));
}

VCL_VOID
vmod_set_add(VRT_CTX, struct vmod_selector_set *set, VCL_STRING member,
	     VCL_STRING string, VCL_STRING regex, VCL_BACKEND backend)
{
	struct entry *entry;
	unsigned n;
	vre_t *re = NULL;
	const char *error;
	int erroffset;
	char **members;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);
	CHECK_OBJ_ORNULL(backend, DIRECTOR_MAGIC);

	if ((ctx->method & VCL_MET_INIT) == 0) {
		VFAIL(ctx, "%s.add() may only be called in vcl_init",
		      set->vcl_name);
		return;
	}

	if (member == NULL) {
		VFAIL(ctx, "%s.add(): string to be added is NULL",
		      set->vcl_name);
		return;
	}

	set->nmembers++;
	n = set->nmembers;
	set->members = realloc(set->members, n * sizeof(VCL_STRING));
	AN(set->members);
	set->members[n - 1] = strdup(member);
	AN(set->members[n - 1]);
	members = set->members;

	if (!set->case_sensitive) {
		set->lomembers = realloc(set->members, n * sizeof(VCL_STRING));
		AN(set->lomembers);
		set->lomembers[n - 1] = strdup(member);
		AN(set->lomembers[n - 1]);
		for (char *m = set->lomembers[n-1]; *m; m++)
			*m = tolower(*m);
		members = set->lomembers;
	}

	errno = 0;
	if (PT_Insert(&set->origo, n - 1, members) != 0) {
		if (errno == EINVAL)
			VFAIL(ctx, "%s.add(): \"%s\" added more than once",
			      set->vcl_name, member);
		else
			VFAIL(ctx, "%s.add(\"%s\") failed: %s", set->vcl_name,
			      member, strerror(errno));
		return;
	}

	if (regex != NULL) {
		/* XXX expose VRE options */
		re = VRE_compile(regex, 0, &error, &erroffset);
		if (re == NULL) {
			VFAIL(ctx, "%s.add(): cannot compile regular expression"
			      " '%s': %s at offset %d", set->vcl_name, regex,
			      error, erroffset);
			return;
		}
	}

	if (string == NULL && re == NULL && backend == NULL)
		return;

	set->nentries = n;
	set->table = realloc(set->table, n * sizeof(struct entry *));
	AN(set->table);
	ALLOC_OBJ(entry, VMOD_SELECTOR_ENTRY_MAGIC);
	AN(entry);
	if (string != NULL) {
		entry->string = strdup(string);
		set_added(set, n - 1, STRING);
	}
	if (re != NULL) {
		entry->re = re;
		set_added(set, n - 1, REGEX);
	}
	if (backend != NULL) {
		entry->backend = backend;
		set_added(set, n - 1, BACKEND);
	}
	set->table[n - 1] = entry;
}

static struct match_data *
get_match_data(VRT_CTX, struct vmod_selector_set *set, const char *method)
{
	struct vmod_priv *task;
	struct match_data *match;

	task = VRT_priv_task(ctx, set);
	AN(task);
	if (task->priv == NULL) {
		if ((task->priv = WS_Alloc(ctx->ws, sizeof(*match))) == NULL) {
			VERRNOMEM(ctx, "Allocating match data in %s.%s()",
				  set->vcl_name, method);
			return (NULL);
		}
		task->len = sizeof(*match);
		task->free = NULL;
		match = (struct match_data *)task->priv;
		match->magic = MATCH_DATA_MAGIC;
	}
	else {
		WS_Assert_Allocated(ctx->ws, task->priv, sizeof(*match));
		CAST_OBJ(match, task->priv, MATCH_DATA_MAGIC);
	}
	return match;
}

VCL_BOOL
vmod_set_match(VRT_CTX, struct vmod_selector_set *set, VCL_STRING subject)
{
	unsigned idx;
	struct match_data *match;
	char **members;
	const char *subj;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);
	AN(set->origo);

	if (subject == NULL) {
		VERR(ctx, "%s.match(): subject string is NULL", set->vcl_name);
		return (0);
	}
	if (set->nmembers == 0) {
		VERR(ctx, "%s.match(): no entries were added", set->vcl_name);
		return (0);
	}

	members = set->members;
	subj = subject;
	if (!set->case_sensitive) {
		char *copy;

		if ((copy = WS_Copy(ctx->ws, subject, -1)) == NULL) {
			VERRNOMEM(ctx, "%s.match(): copying subject for "
				  "case-insensitive match", set->vcl_name);
			return (0);
		}
		for (char *c = copy; *c; c++)
			*c = tolower(*c);
		subj = copy;
		members = set->lomembers;
	}
	AN(members);

	match = get_match_data(ctx, set, "match");
	if ((idx = PT_Lookup(set->origo, members, subj)) == UINT_MAX) {
		match->n = 0;
		return (0);
	}

	if ((match->indices = WS_Alloc(ctx->ws, sizeof(unsigned))) == NULL) {
		VERRNOMEM(ctx, "Reserving space for index in "
			  "%s.match(\"%.40s\")", set->vcl_name, subject);
		return (0);
	}
	*match->indices = idx;
	match->n = 1;
	match->exact = idx;
	return (1);
}

VCL_BOOL
vmod_set_hasprefix(VRT_CTX, struct vmod_selector_set *set, VCL_STRING subject)
{
	struct match_data *match;
	char **members;
	const char *subj;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);
	AN(set->origo);

	if (subject == NULL) {
		VERR(ctx, "%s.hasprefix(): subject string is NULL",
		     set->vcl_name);
		return (0);
	}
	if (set->nmembers == 0) {
		VERR(ctx, "%s.hasprefix(): no entries were added",
		     set->vcl_name);
		return (0);
	}

	members = set->members;
	subj = subject;
	if (!set->case_sensitive) {
		char *copy;

		if ((copy = WS_Copy(ctx->ws, subject, -1)) == NULL) {
			VERRNOMEM(ctx, "%s.hasprefix(): copying subject for "
				  "case-insensitive match", set->vcl_name);
			return (0);
		}
		for (char *c = copy; *c; c++)
			*c = tolower(*c);
		subj = copy;
		members = set->lomembers;
	}
	AN(members);

	match = get_match_data(ctx, set, "hasprefix");
	if ((match->limit = WS_ReserveLumps(ctx->ws, sizeof(unsigned))) == 0) {
		VERRNOMEM(ctx, "Reserving index array in "
			  "%s.hasprefix(\"%.40s\")", set->vcl_name, subject);
		return (0);
	}
	match->indices = (unsigned *)WS_Front(ctx->ws);

	if (PT_Prefixes(set->origo, members, subj, match) != 0) {
		VERRNOMEM(ctx, "Adding indices in %s.hasprefix(\"%.40s\")",
			  set->vcl_name, subject);
		WS_Release(ctx->ws, 0);
		return (0);
	}

	WS_Release(ctx->ws, match->n * sizeof(unsigned));
	return (match->n > 0);
}

static struct match_data *
get_existing_match_data(VRT_CTX,
			const struct vmod_selector_set * const restrict set,
			const char * const restrict method)
{
	struct vmod_priv *task;
	struct match_data *match;

	task = VRT_priv_task(ctx, set);
	AN(task);
	if (task->priv == NULL) {
		VERR(ctx, "%s.%s() called without prior match", set->vcl_name,
		     method);
		return (NULL);
	}
	WS_Assert_Allocated(ctx->ws, task->priv, sizeof(*match));
	CAST_OBJ(match, task->priv, MATCH_DATA_MAGIC);
	return match;
}

VCL_INT
vmod_set_nmatches(VRT_CTX, struct vmod_selector_set *set)
{
	struct match_data *match;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	match = get_existing_match_data(ctx, set, "nmatches");
	if (match == NULL)
		return (0);
	return (match->n);
}

VCL_BOOL
vmod_set_matched(VRT_CTX, struct vmod_selector_set *set, VCL_INT idx)
{
	struct match_data *match;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	if (idx < 1 || idx > set->nmembers) {
		VERR(ctx, "%s.matched(%ld) out of range (%d members)",
		     set->vcl_name, idx, set->nmembers);
		return (0);
	}

	match = get_existing_match_data(ctx, set, "matched");
	if (match == NULL || match->n == 0)
		return (0);

	AN(match->indices);
	WS_Assert_Allocated(ctx->ws, match->indices,
			    match->n * sizeof(unsigned));
	/* XXX search algorithm? */
	idx--;
	for (unsigned i = 0; i < match->n; i++)
		if (match->indices[i] == idx)
			return (1);
	return (0);
}

static unsigned
select(VRT_CTX, const struct match_data * const restrict match,
       const char * const restrict obj, VCL_ENUM const restrict selects,
       const char * const restrict method)
{
	if (selects == vmod_enum_EXACT) {
		if (match->exact == UINT_MAX)
			VERR(ctx, "%s.%s(select=EXACT): "
			     "no element matched exactly", obj, method);
		return match->exact;
	}
	if (match->n == 1)
		return match->indices[0];

	if (selects == vmod_enum_UNIQUE) {
		VERR(ctx, "%s.%s(select=UNIQUE): %d elements were matched",
		     obj, method, match->n);
		return (UINT_MAX);
	}
	if (selects == vmod_enum_FIRST)
		return match->min;
	if (selects == vmod_enum_LAST)
		return match->max;
	if (selects == vmod_enum_SHORTEST)
		return match->indices[0];
	if (selects == vmod_enum_LONGEST)
		return match->indices[match->n - 1];
	WRONG("illegal select enum");
	return (UINT_MAX);
}

VCL_INT
vmod_set_which(VRT_CTX, struct vmod_selector_set *set, VCL_ENUM selects)
{
	struct match_data *match;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	match = get_existing_match_data(ctx, set, "which");
	if (match == NULL || match->n == 0)
		return (0);
	return (select(ctx, match, set->vcl_name, selects, "which") + 1);
}

static unsigned
get_idx(VRT_CTX, VCL_INT n, const struct vmod_selector_set * const restrict set,
	const char * const restrict method, VCL_ENUM const restrict selects)
{
	struct match_data *match;

	if (n > 0) {
		if (n > set->nmembers) {
			VERR(ctx, "%s.%s(%ld): set has %d elements",
			     set->vcl_name, method, n, set->nmembers);
			return (UINT_MAX);
		}
		return (n - 1);
	}

	match = get_existing_match_data(ctx, set, method);
	if (match == NULL || match->n == 0)
		return (UINT_MAX);
	return (select(ctx, match, set->vcl_name, selects, method));
}

VCL_STRING
vmod_set_element(VRT_CTX, struct vmod_selector_set *set, VCL_INT n,
		 VCL_ENUM selects)
{
	unsigned idx;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	idx = get_idx(ctx, n, set, "element", selects);
	if (idx == UINT_MAX)
		return (NULL);
	return (set->members[idx]);
}

static inline int
check_added(VRT_CTX, const struct vmod_selector_set * const restrict set,
	    unsigned idx, enum bitmap_e bitmap,
	    const char * const restrict method,
	    const char * const restrict type)
{
	if (!is_added(set, idx, bitmap)) {
		VERR(ctx, "%s.%s(): %s not added for element %u", set->vcl_name,
		     method, type, idx + 1);
		return (0);
	}
	return (1);
}

VCL_BACKEND
vmod_set_backend(VRT_CTX, struct vmod_selector_set *set, VCL_INT n,
		 VCL_ENUM selects)
{
	unsigned idx;
	VCL_BACKEND b;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	idx = get_idx(ctx, n, set, "backend", selects);
	if (idx == UINT_MAX)
		return (NULL);
	if (!check_added(ctx, set, idx, BACKEND, "backend", "backend"))
		return (NULL);

	b = set->table[idx]->backend;
	CHECK_OBJ_NOTNULL(b, DIRECTOR_MAGIC);
	return (b);
}

VCL_STRING
vmod_set_string(VRT_CTX, struct vmod_selector_set * set, VCL_INT n,
		VCL_ENUM selects)
{
	unsigned idx;
	VCL_STRING s;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	idx = get_idx(ctx, n, set, "string", selects);
	if (idx == UINT_MAX)
		return (NULL);
	if (!check_added(ctx, set, idx, STRING, "string", "string"))
		return (NULL);

	s = set->table[idx]->string;
	AN(s);
	return (s);
}

VCL_BOOL
vmod_set_re_match(VRT_CTX, struct vmod_selector_set *set, VCL_STRING subject,
		  VCL_INT n, VCL_ENUM selects)
{
	unsigned idx;
	vre_t *re;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	idx = get_idx(ctx, n, set, "re_match", selects);
	if (idx == UINT_MAX)
		return (0);
	if (!check_added(ctx, set, idx, REGEX, "re_match", "regex"))
		return (0);

	re = set->table[idx]->re;
	AN(re);
	return (VRT_re_match(ctx, subject, re));
}

VCL_STRING
vmod_set_debug(VRT_CTX, struct vmod_selector_set *set)
{
	struct vsb *sb;
	VCL_STRING output;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_SELECTOR_SET_MAGIC);

	sb = PT_Dump(set->origo, set->members);
	CHECK_OBJ_NOTNULL(sb, VSB_MAGIC);
	output = WS_Copy(ctx->ws, VSB_data(sb), -1);
	VSB_destroy(&sb);
	if (output == NULL)
		VFAIL(ctx, "%s.debug(): out of workspace", set->vcl_name);
	return output;
}

VCL_STRING
vmod_version(VRT_CTX)
{
	(void) ctx;
	return VERSION;
}
